##############################################################################
#
#  List the test cases that are to be disabled temporarily.
#
#  Separate the test case name and the comment with ':'.
#
#    <testcasename> : BUG#<xxxx> <date disabled> <disabler> <comment>
#
#  Do not use any TAB characters for whitespace.
#
##############################################################################

rpl_spec_variables        : BUG#11755836 2009-10-27 jasonh rpl_spec_variables fails on PB2 hpux
rpl_get_master_version_and_clock : Bug#11766137 Jan 05 2011 joro Valgrind warnings rpl_get_master_version_and_clock
rpl_partition_archive     : MDEV-5077 2013-09-27 svoj Cannot exchange partition with archive table
rpl_row_binlog_max_cache_size : MDEV-11092
rpl_blackhole             : MDEV-11094
rpl_row_mysqlbinlog       : MDEV-11095
rpl_row_index_choice      : MDEV-11666
